(*
Isabelle/HOL source for lectures of Set Theory
created by Yongwang ZHAO
School of Computer Science and Engineering, Beihang University, China
zhaoyw@buaa.edu.cn
*)


theory Function
imports Main Real_Vector_Spaces NthRoot

begin

chapter {* Chapter 7. Function *}

section {* starter *}

(* domain and range of a function should not be empty*)
typedef myint = "{1::int,2}"
  by force

consts afunc :: "int \<Rightarrow> myint" 

definition afunc2 :: "int \<Rightarrow> int"
  where "afunc2 \<equiv> \<lambda>x. x + 1"

definition afunc3 :: "int \<Rightarrow> int"
  where "afunc3 \<equiv> \<lambda>_. 0"

lemma "afunc3 x = (0::int)"
  unfolding afunc3_def by simp

section {* section 7.1: concepts *}

subsection {* definition 7.1, 7.2 *}


type_synonym ('a, 'b) Relation = "('a \<times> 'b) set" 

definition func :: "('a, 'b) Relation \<Rightarrow> 'a set \<Rightarrow> 'b set \<Rightarrow> bool" ("_ : _ \<rightarrow> _" [80,80,80] 81)
  where "func f X Y \<equiv> f \<subseteq> X \<times> Y \<and> X \<noteq> {} \<and> Y \<noteq> {} \<and> Domain f = X
                        \<and> (\<forall>x y1 y2. (x, y1)\<in>f \<and> (x, y2)\<in>f \<longrightarrow> y1 = y2)" 

lemma func_total: "f : X \<rightarrow> Y \<Longrightarrow> \<forall>x. x\<in>X \<longrightarrow>  (\<exists>y. (x, y) \<in> f)" 
  unfolding func_def by blast

definition val :: "('a, 'b) Relation \<Rightarrow> 'a \<Rightarrow> 'b" ("_\<langle>_\<rangle>" [80,80] 80) 
  where "f\<langle>x\<rangle> \<equiv> (SOME y. (x, y) \<in> f)" 

definition val_valid1 :: "('a, 'b) Relation \<Rightarrow> 'a \<Rightarrow> 'b \<Rightarrow> bool" ("_\<langle>_\<rangle> \<doteq> _" [80,80] 80) 
  where "(f\<langle>x\<rangle> \<doteq> y) \<equiv> (x, y) \<in> f" 

definition val_valid2 :: "'b \<Rightarrow> ('a, 'b) Relation \<Rightarrow> 'a \<Rightarrow> bool" ("_ \<doteq> _\<langle>_\<rangle>" [80,80] 80) 
  where "(y \<doteq> f\<langle>x\<rangle>) \<equiv> (x, y) \<in> f" 

definition val_rel1 :: "('a, 'b) Relation \<Rightarrow> 'a \<Rightarrow> ('b \<Rightarrow> 'b \<Rightarrow> bool) 
                          \<Rightarrow> 'b \<Rightarrow> bool" ("_\<langle>_\<rangle> _ _" [80,80] 80) 
  where "(f\<langle>x\<rangle> r y) \<equiv> (\<exists>y1. (x, y1) \<in> f \<and> r y1 y)" 

subsection {* example 7.1 *}

definition f71:: "int \<Rightarrow> int"
  where "f71 x \<equiv> (if x mod 2 \<noteq> 0 then 1 else x div 2)"

subsection {* example 7.2 *}

definition "f1 \<equiv> {(x::real, y). y = x * x}"

lemma "f1 : \<real> \<rightarrow> \<real>" 
  unfolding func_def f1_def
    apply(rule conjI)
    unfolding Reals_def by force+

subsection {* definition 7.3 *}

definition func_restri :: "('a, 'b) Relation \<Rightarrow> 'a set \<Rightarrow> ('a, 'b) Relation" ("_ | _")
  where "f | A \<equiv> {(x,y). (x,y)\<in>f \<and> x\<in>A}"

lemma "(f : X \<rightarrow> Y) \<and> A \<subseteq> X \<and> A \<noteq> {} \<Longrightarrow> (f | A) : (X \<inter> A) \<rightarrow> Y" 
  unfolding func_restri_def func_def by fastforce

lemma "(f : X \<rightarrow> Y) \<and> A \<subseteq> X \<Longrightarrow> Domain (f | A) = A" 
  unfolding func_restri_def func_def by force

lemma "(f : X \<rightarrow> Y) \<and> A \<subseteq> X \<Longrightarrow> \<forall>a\<in>A. (f | A)\<langle>a\<rangle> = f\<langle>a\<rangle>" 
  unfolding func_restri_def func_def  snd_conv val_valid1_def val_def by force

subsection {* definition 7.4 *}

definition func_image :: "('a, 'b) Relation \<Rightarrow> 'a set \<Rightarrow> 'b set" ("_ [_]")
  where "func_image f X \<equiv> {y. \<exists>x\<in>X. (x, y)\<in>f}"

lemma "(f : X \<rightarrow> Y) \<Longrightarrow> f[X] = Range f" 
  unfolding func_image_def func_def Range_def by force

definition Ff :: "('a, 'b) Relation \<Rightarrow> 'a set \<Rightarrow> 'b set \<Rightarrow> ('a set \<times> 'b set) set"
  where "Ff f X Y \<equiv> {(A,B). A \<subseteq> X \<and> B = {y. \<exists>x\<in>X. f\<langle>x\<rangle> \<doteq> y}}"

lemma "f : X \<rightarrow> Y \<Longrightarrow> (Ff f X Y) : (Pow X) \<rightarrow> (Pow Y)"
  unfolding Ff_def val_valid1_def Pow_def func_def by blast

subsection {* example 7.10 *}

definition "f710 \<equiv> {(0::int,''b''),(1,''a''),(2,''c''),(3,''b'')}"

lemma "f710 : {0::int,1,2,3} \<rightarrow> {''a'',''b'',''c''}"
  unfolding func_def f710_def by force

lemma "f710 [{0,1,2,3}] = {''a'',''b'',''c''}"
  unfolding f710_def func_image_def by force

lemma "f710 [{2,3}] = {''b'',''c''}"
  unfolding f710_def func_image_def by force

lemma "f710 [{0}] = {''b''}"
  unfolding f710_def func_image_def by force

lemma "f710 [{0,3}] = {''b''}"
  unfolding f710_def func_image_def by force

lemma "f710 [{}] = {}"
  unfolding f710_def func_image_def by force

subsection {* definition 7.5 *}

definition partial_func :: "('a, 'b) Relation \<Rightarrow> 'a set \<Rightarrow> 'b set \<Rightarrow> bool" ("_ : _ \<circ>\<rightarrow> _" [80,80,80] 81)
  where "partial_func f X Y \<equiv> f \<subseteq> X \<times> Y \<and> X \<noteq> {} \<and> Y \<noteq> {} 
                        \<and> (\<forall>x y1 y2. (x, y1)\<in>f \<and> (x, y2)\<in>f \<longrightarrow> y1 = y2)" 

lemma "f : X \<rightarrow> Y \<Longrightarrow> f : X \<circ>\<rightarrow> Y"
  unfolding func_def partial_func_def by blast

subsection {* example 7.12 *}

definition "f2 \<equiv> {(x::real, y). y = sqrt x}" 

lemma "f2 : \<real> \<circ>\<rightarrow> \<real>" 
  unfolding partial_func_def f2_def 
    apply(rule conjI)
    unfolding Reals_def by force+

definition FXY :: "'a set \<Rightarrow> 'b set \<Rightarrow> ('a \<times> 'b) set set" 
  where "FXY X Y \<equiv> {f. f : X \<rightarrow> Y}"

section {* section 7.2 composition of functions *}

subsection {* definition 7.6 *}

definition func_comp :: "('b, 'c) Relation \<Rightarrow> ('a, 'b) Relation \<Rightarrow> ('a, 'c) Relation" ("_ \<diamondop> _" [80,80] 81)
  where "g \<diamondop> f \<equiv> {(x, z). \<exists>y. (x, y) \<in> f \<and> (y, z) \<in> g}" 

subsection {* theorem 7.1 *}

(* f: X \<longrightarrow> Y, g: W \<longrightarrow> Z, we dont require Y = W, we only require Y \<subseteq> W *)
lemma thm7_1_1: "f : X \<rightarrow> Y \<and> g : Y \<rightarrow> Z \<Longrightarrow> (g \<diamondop> f) : X \<rightarrow> Z"
  unfolding func_def func_comp_def
    apply(rule conjI) 
      apply blast apply(rule conjI)
        apply simp apply(rule conjI)
          apply simp apply(rule conjI)
            by blast+
    
lemma thm7_1_2_: "f : X \<rightarrow> Y \<and> g : Y \<rightarrow> Z \<Longrightarrow> \<forall>y.  (y \<doteq> (g \<diamondop> f)\<langle>x\<rangle>) = (\<exists>z. z \<doteq> f\<langle>x\<rangle> \<and> y \<doteq> g\<langle>z\<rangle>)"
  unfolding func_def func_comp_def val_def val_valid2_def by fastforce
    
subsection {* theorem 7.2 *}
definition "Idx X \<equiv> {p. \<exists>x\<in>X. p = (x,x)}" 

lemma thm7_2_1: "f : X \<rightarrow> Y  \<Longrightarrow> f = f \<diamondop> (Idx X)" 
  unfolding func_def Idx_def func_comp_def by force

lemma thm7_2_2: "f : X \<rightarrow> Y  \<Longrightarrow> f = (Idx Y) \<diamondop> f"
  unfolding func_def Idx_def func_comp_def by force

subsection {* theorem 7.3 *}

lemma thm7_3: "h \<diamondop> (g \<diamondop> f) = (h \<diamondop> g) \<diamondop> f" 
  unfolding func_comp_def by force

subsection {* definition 7.7 *}

primrec F_n :: "('a, 'a) Relation \<Rightarrow> nat \<Rightarrow> ('a, 'a) Relation" ("_\<^sup>_")
  where F_n_zero: "f\<^sup>0 = Idx (Domain f)" |
        F_n_suc:  "F_n f (Suc n) = (f\<^sup>n) \<diamondop> f"

subsection {* example 7.14 *}

definition "f3 \<equiv> {(1::int, 2), (2, 3), (3, 1)}"
definition "h3 \<equiv> {(1::int, 1), (2, 2), (3, 1)}"
definition "g3 \<equiv> {(1::int, 2), (2, 1), (3, 3)}"

lemma "f3 : {1,2,3} \<rightarrow> {1,2,3}"
  unfolding func_def f3_def by force

lemma "h3 : {1,2,3} \<rightarrow> {1,2,3}"
  unfolding func_def h3_def by force

lemma "g3 : {1,2,3} \<rightarrow> {1,2,3}"
  unfolding func_def g3_def by force

lemma "f3 \<diamondop> g3 = {(1, 3), (2, 2), (3, 1)}" 
  unfolding func_comp_def f3_def g3_def by fastforce

lemma "g3 \<diamondop> f3 = {(1, 1), (2, 3), (3 ,2)}"
  unfolding func_comp_def f3_def g3_def by fastforce

lemma "f3 \<diamondop> (h3 \<diamondop> g3) = {(1, 3), (2, 2), (3, 2)}"
  unfolding func_comp_def f3_def g3_def h3_def by fastforce

subsection {* example 7.15 *}

definition "f4 = {(x::real, y). y = x + 2}"
definition "g4 = {(x::real, y). y = x - 2}"
definition "h4 = {(x::real, y). y = x * 3}"

lemma "f4 : \<real> \<rightarrow> \<real> "
  unfolding func_def f4_def Reals_def by force

lemma "g4 : \<real> \<rightarrow> \<real> "
  unfolding func_def g4_def Reals_def by force

lemma "g4 : \<real> \<rightarrow> \<real> "
  unfolding func_def g4_def Reals_def by force

lemma lm7151: "g4 \<diamondop> f4 = {(x::real, y). x = y}"
  unfolding func_comp_def f4_def g4_def by fastforce

lemma "g4 \<diamondop> f4 : \<real> \<rightarrow> \<real> "
  unfolding func_def Reals_def using lm7151 by force

lemma lm7152: "f4 \<diamondop> g4 = {(x::real, y). x = y}"
  unfolding func_comp_def f4_def g4_def by fastforce

lemma "f4 \<diamondop> g4 : \<real> \<rightarrow> \<real> "
  unfolding func_def Reals_def using lm7152 by force

lemma lm7153: "f4 \<diamondop> f4 = {(x::real, y). y = x + 4}"
  unfolding func_comp_def f4_def by fastforce

lemma "f4 \<diamondop> f4 : \<real> \<rightarrow> \<real> "
  unfolding func_def Reals_def using lm7153 by force

lemma lm7154: "g4 \<diamondop> g4 = {(x::real, y). y = x - 4}"
  unfolding func_comp_def g4_def by fastforce

lemma "g4 \<diamondop> g4 : \<real> \<rightarrow> \<real> "
  unfolding func_def Reals_def using lm7154 by force

lemma lm7155: "f4 \<diamondop> h4 = {(x::real, y). y = 3 * x + 2}"
  unfolding func_comp_def f4_def h4_def by fastforce

lemma "f4 \<diamondop> h4 : \<real> \<rightarrow> \<real> "
  unfolding func_def Reals_def using lm7155 by force

lemma lm7156: "h4 \<diamondop> g4 = {(x::real, y). y = 3 * x - 6}"
  unfolding func_comp_def g4_def h4_def by fastforce

lemma "h4 \<diamondop> g4 : \<real> \<rightarrow> \<real> "
  unfolding func_def Reals_def using lm7156 by force

lemma lm7157: "h4 \<diamondop> f4 = {(x::real, y). y = 3 * x + 6}"
  unfolding func_comp_def f4_def h4_def by fastforce

lemma "h4 \<diamondop> f4 : \<real> \<rightarrow> \<real> "
  unfolding func_def Reals_def using lm7157 by force

subsection {* example 7.16 *}
definition "g5 \<equiv> {(x::nat,y). y = x * 2}"
definition "f5 \<equiv> {(x::nat,y). (if x mod 2 = 0 then x = y * 2 else y = 0)}"

lemma lm7161: "f5 \<diamondop> g5 = {(x, y). y = x}"
  unfolding func_comp_def f5_def g5_def by fastforce

lemma "g5 \<diamondop> f5 = {(x, y). (x mod 2 = 0 \<longrightarrow> y = x) 
                     \<and> (x mod 2 \<noteq> 0 \<longrightarrow> y = 0)}"
  unfolding func_comp_def f5_def g5_def 
      apply(simp add:set_eq_iff)
      apply(rule allI)+
      apply(case_tac "a mod 2 = 0")
      by force+

section {* section 7.3 special functions *}

subsection {* defintion 7.8 *}

definition surjective :: "('a, 'b) Relation \<Rightarrow> 'a set \<Rightarrow> 'b set \<Rightarrow> bool" ("_ : _ \<rightarrow>\<bullet> _" [80,80,80] 81)
  where "surjective f X Y\<equiv> f : X \<rightarrow> Y \<and> (\<forall>y\<in>Y. (\<exists>x\<in>X. y \<doteq> f\<langle>x\<rangle>))"

definition injective :: "('a, 'b) Relation \<Rightarrow> 'a set \<Rightarrow> 'b set \<Rightarrow> bool" ("_ : _ \<rightarrow>- _" [80,80,80] 81)
  where "injective f X Y \<equiv> f : X \<rightarrow> Y \<and> (\<forall>x1 x2 y. (x1,y)\<in>f \<and> (x2,y)\<in>f \<longrightarrow> x1 = x2)"

definition bijective :: "('a, 'b) Relation \<Rightarrow> 'a set \<Rightarrow> 'b set \<Rightarrow> bool" ("_ : _ \<leftrightarrow> _" [80,80,80] 81)
  where "bijective f X Y \<equiv> surjective f X Y \<and> injective f X Y"

subsection {* example 7.17 *}

definition "f6 \<equiv> {(1, 0), (2, 0)}"

lemma exam7_17_1: "f6 : {1,2} \<rightarrow>\<bullet> {0}"
  unfolding func_def f6_def surjective_def val_valid2_def by fastforce

definition "f7 \<equiv> {(''a'', 2::int), (''b'', 6)}"

lemma exam7_17_2: "f7 : {''a'', ''b''} \<rightarrow>- {2::int,4,6}"
  unfolding func_def f7_def injective_def val_valid2_def by fastforce

definition "f8 \<equiv> {(x::nat, y). y = x * 2}"

lemma "f8 : \<nat> \<rightarrow>- \<nat>"
  unfolding func_def f8_def Nats_def injective_def by fastforce

definition "f9 \<equiv> {(x::int, y::int). y = x + 1}"

lemma "f9 : \<int> \<leftrightarrow> \<int>" 
  unfolding bijective_def surjective_def injective_def 
    func_def f9_def Ints_def val_valid2_def
    apply(rule conjI)+ 
      apply force+
      prefer 2 apply force
      apply auto[1]
        proof -
        {
          fix y
          have "(y::int) = (y - 1) + 1" by auto
          then show "\<exists>x. (y::int) = x + 1" by blast
        }
      qed


subsection {* converse of relation *}

definition Converse :: "('a, 'b) Relation \<Rightarrow> ('b, 'a) Relation" ("(_\<^sup>-\<^sup>1)" [1000] 999)
  where "Converse f \<equiv> {(y,x). (x,y)\<in>f}"

lemma Conv_Conv: "(r\<^sup>-\<^sup>1)\<^sup>-\<^sup>1 = r"
  unfolding Converse_def by force

lemma Converse_iff:
  "(x, y) \<in> r\<^sup>-\<^sup>1 \<longleftrightarrow> (y, x) \<in> r"
  unfolding Converse_def by auto

subsection {* theorem 7.4 *}

lemma thm7_4: "f : X \<leftrightarrow> Y \<Longrightarrow>f\<^sup>-\<^sup>1 : Y \<leftrightarrow> X"
  proof -
    assume p0: "f : X \<leftrightarrow> Y"
    then have "f : X \<rightarrow>\<bullet> Y" by (simp add:bijective_def)
    then have a0: "\<forall>y\<in>Y. (\<exists>x\<in>X. y \<doteq> f\<langle>x\<rangle>)" by (simp add:surjective_def)
    from p0 have a1: "\<forall>x1 x2 y. (x1,y)\<in> f \<and> (x2,y)\<in> f \<longrightarrow> x1 = x2" 
       by (simp add:bijective_def injective_def)
    from p0 have a2: "f : X \<rightarrow> Y" by (simp add:bijective_def surjective_def)
    have g1: "f\<^sup>-\<^sup>1 : Y \<rightarrow> X"
      unfolding func_def apply(rule conjI)
        using a2 unfolding func_def Converse_def apply fastforce
        apply(rule conjI)
          using a2 unfolding func_def apply simp
          apply(rule conjI)
          using a2 unfolding func_def apply simp
            apply(rule conjI)
              using a0 a2 unfolding func_def val_valid2_def apply force
              apply(rule allI)+
              using a1 a2 unfolding val_valid2_def Converse_def func_def by blast
              
    from g1 a2 have g2: "\<forall>y\<in>X. (\<exists>x\<in>Y. y \<doteq> f\<^sup>-\<^sup>1\<langle>x\<rangle>)" 
      unfolding func_def Converse_def val_valid2_def by fastforce
    
    from a0 a1 a2 have g3: "\<forall>x1 x2 y. (x1,y)\<in>f\<^sup>-\<^sup>1 \<and> (x2,y)\<in>f\<^sup>-\<^sup>1 \<longrightarrow> x1 = x2" 
      unfolding func_def val_valid2_def Converse_def by blast
    
    from g1 g2 g3 show ?thesis 
      unfolding bijective_def surjective_def injective_def by metis
  qed

subsection {* example 7.18 *}

definition "f10 \<equiv> {(1::int,''p''),(2,''q''),(3,''q'')}"

lemma "f10 : {1,2,3} \<rightarrow> {''p'',''q'',''r''}"
  unfolding func_def f10_def by force

value "f10\<^sup>-\<^sup>1"

subsection {* example 7.19 *}

definition "f11 \<equiv> {(x::real,y::real). y = x\<^sup>2}"

lemma "f11 : \<real> \<rightarrow> \<real>"
  unfolding func_def f11_def Reals_def by force

subsection {* example 7.20 *}

definition "f12 \<equiv> {(x::real,y::real). y = x + 2}"

lemma exm720: "f12 : \<real> \<leftrightarrow> \<real>"
  unfolding bijective_def surjective_def injective_def func_def f12_def Reals_def
    apply(rule conjI)+
    apply force+
    prefer 2
    apply(rule conjI)+
    apply force+
    unfolding val_valid2_def 
    proof
      fix y
      assume a0: "y \<in> range of_real"
      have "(y-2,y)\<in> {(x, y). y = x + 2}" by force
      then show "\<exists>x\<in>range of_real. (x, y) \<in> {(x, y). y = x + 2} "
        by (metis Reals_def Reals_diff Reals_numeral a0) 
    qed

lemma "f12\<^sup>-\<^sup>1={(x::real,y::real). y = x - 2}"
  unfolding Converse_def f12_def by force

lemma "f12\<^sup>-\<^sup>1 : \<real> \<leftrightarrow> \<real>"
  using exm720 thm7_4 by blast

subsection {* example 7.21 *}

definition "f13 \<equiv> {(0::int,''p''),(1,''q'')}"

lemma "f13 : {0,1} \<rightarrow> {''p'',''q'',''r'',''s''}"
  unfolding func_def f13_def by force

value "f13\<^sup>-\<^sup>1"

lemma "f13\<^sup>-\<^sup>1 : {''p'',''q'',''r'',''s''} \<circ>\<rightarrow> {0,1}"
  unfolding partial_func_def Converse_def f13_def by force

subsection {* theorem 7.5 *}

lemma conv_f_val: "f\<langle>x\<rangle> \<doteq> y \<Longrightarrow> f\<^sup>-\<^sup>1\<langle>y\<rangle> \<doteq> x" 
  unfolding val_valid2_def val_valid1_def 
    using Converse_iff by fastforce

lemma thm7_5_1: "f : X \<leftrightarrow> Y \<Longrightarrow> f\<^sup>-\<^sup>1 \<diamondop> f = Idx X"
  proof -
    assume p0: "f : X \<leftrightarrow> Y"
    then have p1: "f\<^sup>-\<^sup>1 : Y \<leftrightarrow> X" using thm7_4 by blast
    
    {
      fix x y
      assume "(x,y)\<in> (f\<^sup>-\<^sup>1 \<diamondop> f)"
      then have "\<exists>z. (x, z) \<in> f \<and> (z, y) \<in> f\<^sup>-\<^sup>1" 
        unfolding func_comp_def by simp
      then obtain z where "(x, z) \<in> f \<and> (z, y) \<in> f\<^sup>-\<^sup>1" by auto
      then have a1: "(x, z) \<in> f \<and> (y, z) \<in> f" 
        using Converse_iff by fastforce
      with p0 have "x\<in>X \<and> y\<in>X \<and> z\<in>Y" 
        unfolding bijective_def injective_def func_def by blast
      with p0 a1 have "x = y" 
        unfolding bijective_def injective_def func_def by blast
      with p0 a1 have "(x,y)\<in>Idx X" 
        unfolding Idx_def bijective_def injective_def func_def by blast
    }
    moreover
    {
      fix x y
      assume "(x,y)\<in>Idx X"
      then have a0: "x\<in>X \<and> y\<in>X \<and> x = y" 
        unfolding Idx_def by blast
      with p0 have "\<exists>z\<in>Y. (x,z)\<in> f" 
        unfolding bijective_def injective_def func_def by blast
      then obtain z where a1: "z\<in>Y \<and> (x,z)\<in> f" by auto
      then have "(z,x)\<in>f\<^sup>-\<^sup>1" using Converse_iff by fastforce
      with a1 have "\<exists>y. (x, y) \<in> f \<and> (y, x) \<in> f\<^sup>-\<^sup>1" by auto
      
      with a0 have "(x,y)\<in>(f\<^sup>-\<^sup>1 \<diamondop> f)" unfolding func_comp_def by fastforce
    }
    ultimately show ?thesis by auto
  qed


lemma thm7_5_2: "f : X \<leftrightarrow> Y \<Longrightarrow> f \<diamondop> (f\<^sup>-\<^sup>1) = Idx Y" sorry
(* please prove it by youself *)

subsection {* theorem 7.6 *}

lemma thm7_6: "f : X \<leftrightarrow> Y \<and> g : Y \<leftrightarrow> X \<Longrightarrow> (g = f\<^sup>-\<^sup>1) = (g \<diamondop> f = Idx X \<and> f \<diamondop> g = Idx Y)"
  proof 
    assume p0: "f : X \<leftrightarrow> Y \<and> g : Y \<leftrightarrow> X"
      and  p1: "g = f\<^sup>-\<^sup>1"
    then show "g \<diamondop> f = Idx X \<and> f \<diamondop> g = Idx Y" 
      using thm7_5_1 thm7_5_2 by blast 
  next
    assume p0: "f : X \<leftrightarrow> Y \<and> g : Y \<leftrightarrow> X"
      and  p1: "g \<diamondop> f = Idx X \<and> f \<diamondop> g = Idx Y"
    then have "(g \<diamondop> f) \<diamondop> (f\<^sup>-\<^sup>1) = Idx X \<diamondop> (f\<^sup>-\<^sup>1)" by simp
    with p0 have "(g \<diamondop> f) \<diamondop> (f\<^sup>-\<^sup>1) = f\<^sup>-\<^sup>1" 
      using thm7_2_2[of "f\<^sup>-\<^sup>1" Y X] thm7_4[of f X Y] Converse_def 
        unfolding bijective_def surjective_def by force
      
    then have "g \<diamondop> (f \<diamondop> (f\<^sup>-\<^sup>1)) = f\<^sup>-\<^sup>1" using thm7_3[of g f "f\<^sup>-\<^sup>1"] by fastforce
    then have "g \<diamondop> (Idx Y) = f\<^sup>-\<^sup>1" using thm7_5_2[of f] p0 by fastforce
    then show "g = f\<^sup>-\<^sup>1" using p0 thm7_2_1[of] 
      unfolding bijective_def surjective_def by blast
  qed


subsection {* theorem 7.7 *}

lemma thm7_7_1: "f : X \<leftrightarrow> Y \<and> g : Y \<leftrightarrow> Z \<Longrightarrow> (g \<diamondop> f) : X \<leftrightarrow> Z" 
  proof -
    assume p0: "f : X \<leftrightarrow> Y \<and> g : Y \<leftrightarrow> Z"
    
    from p0 have a2: "\<forall>y\<in>Y. (\<exists>x\<in>X. y \<doteq> f\<langle>x\<rangle>)" 
      using bijective_def surjective_def by metis

    from p0 have a3: "\<forall>x1 x2 y. (x1,y)\<in>f \<and> (x2,y)\<in>f \<longrightarrow> x1 = x2"
      by (simp add:bijective_def injective_def func_def)
    
    from p0 have b2: "\<forall>y\<in>Z. (\<exists>x\<in>Y. y \<doteq> g\<langle>x\<rangle>)" 
      using bijective_def surjective_def by metis

    from p0 have b3: "\<forall>x1 x2 y. (x1,y)\<in>g \<and> (x2,y)\<in>g \<longrightarrow> x1 = x2"
      by (simp add:bijective_def injective_def func_def)

    from p0 have g1: "(g \<diamondop> f) : X \<rightarrow> Z" 
      unfolding bijective_def surjective_def 
        using thm7_1_1 by metis

    let ?gf= "g \<diamondop> f"
    {
      fix x1 x2 y
      assume c0: "(x1,y)\<in>?gf \<and> (x2,y)\<in>?gf"
      with g1 have c01: "x1\<in>X \<and> x2\<in>X \<and> y\<in>Z"
        unfolding func_def by blast
      with p0 have c1: "x1\<in>X" 
        unfolding func_comp_def by fastforce
      moreover
      from c01 p0 have c2: "x2\<in>X" 
        unfolding func_comp_def by fastforce
      then have "(\<exists>z. z \<doteq> f\<langle>x1\<rangle> \<and> y \<doteq> g\<langle>z\<rangle>) \<and> (\<exists>z. z \<doteq> f\<langle>x2\<rangle> \<and> y \<doteq> g\<langle>z\<rangle>)" 
        using c0 p0 thm7_1_2_[of f X Y g Z] thm7_1_2_[of f X Y g Z]
          unfolding val_valid2_def bijective_def surjective_def by blast
      then obtain z1 and z2 where c4: "z1 \<doteq> f\<langle>x1\<rangle> \<and> y \<doteq> g\<langle>z1\<rangle> \<and> z2 \<doteq> f\<langle>x2\<rangle> \<and> y \<doteq> g\<langle>z2\<rangle>" by auto
      with b3 have "z1 = z2" unfolding val_valid2_def by auto
      with a3 c4 have "x1 = x2" unfolding val_valid2_def by auto
    }
    then have g2: "\<forall>x1 x2 y. (x1,y)\<in>?gf \<and> (x2,y)\<in>?gf \<longrightarrow> x1 = x2" by auto

    {
      fix y
      assume c0: "y\<in>Z"
      from b2 c0 have "\<exists>z\<in>Y. y \<doteq> g\<langle>z\<rangle>" by blast
      then obtain z where c2: "z\<in>Y \<and> y \<doteq> g\<langle>z\<rangle>" by auto
      with a2 have "\<exists>x\<in>X. z \<doteq> f\<langle>x\<rangle>" by fastforce
      then obtain x where c3: "x\<in>X \<and> z \<doteq> f\<langle>x\<rangle>" by auto
      
      from p0 g1 c2 c3 have "\<exists>x\<in>X. y \<doteq> ?gf\<langle>x\<rangle>" 
        using thm7_1_2_[of f X Y g Z]
        unfolding bijective_def surjective_def val_valid2_def func_comp_def by fastforce
    }
    then have g3: "\<forall>y\<in>Z. (\<exists>x\<in>X. y \<doteq> ?gf\<langle>x\<rangle>)" by auto

    from g1 g2 g3 show ?thesis by (simp add:bijective_def surjective_def injective_def)
  qed

lemma thm7_7_2: "(g \<diamondop> f)\<^sup>-\<^sup>1 = (f\<^sup>-\<^sup>1) \<diamondop> (g\<^sup>-\<^sup>1)" 
  unfolding func_comp_def Converse_def by fastforce
    
subsection {* F_x *}

definition F :: "'a set \<Rightarrow> ('a,'a) Relation set" 
  where "F X \<equiv> {f. f : X \<leftrightarrow> X}"

lemma fx_1: "f \<in> F X \<and> g \<in> F X \<Longrightarrow> (f \<diamondop> g) \<in> F X \<and> (g \<diamondop> f) \<in> F X"
  unfolding F_def using thm7_7_1  by fastforce

lemma fx_2: "f \<in> F X \<and> g \<in> F X \<and> h \<in> F X
                \<Longrightarrow> (f \<diamondop> g) \<diamondop> h \<in> F X \<and> f \<diamondop> (g \<diamondop> h) \<in> F X"
  using fx_1 by blast

lemma fx_3: "f \<in> F X \<Longrightarrow> f \<diamondop> Idx X = Idx X \<diamondop> f"
  unfolding F_def bijective_def surjective_def 
    using thm7_2_1 thm7_2_2 by blast

lemma fx_4: "f \<in> F X \<Longrightarrow> f\<^sup>-\<^sup>1 \<in> F X \<and> f\<^sup>-\<^sup>1 \<diamondop> f = Idx X"
  using thm7_5_1 thm7_4 unfolding F_def by blast

section {* feature function of sets*}

subsection {* definition 7.11 *}

definition \<Psi> :: "'d set \<Rightarrow> 'd \<Rightarrow> int" 
  where "\<Psi> A x \<equiv> (if x\<in>A then 1 else 0)"  

subsection {* example 7.23 *}

datatype Undergraduate = Female | Male

lemma "{x::Undergraduate. True} = UNIV"
  by force

definition "FemaleSet \<equiv> {x. x = Female}"

lemma "x = Female \<Longrightarrow> \<Psi> FemaleSet x = 1"
  unfolding \<Psi>_def FemaleSet_def by auto

lemma "x = Male \<Longrightarrow> \<Psi> FemaleSet x = 0"
  unfolding \<Psi>_def FemaleSet_def by auto

subsection {* example 7.24 *}

datatype SetA = a | b | c

value "\<Psi> {a} a" 
value "\<Psi> {a} b" 
value "\<Psi> {a} c" 

subsection {* lemmas of feature function *}

lemma featfun_lm1: "(\<forall>x. \<Psi> A x = 0) = (A = {})"
  proof 
    assume a0: "\<forall>x. \<Psi> A x = 0"
    then show "A = {}" unfolding \<Psi>_def by (smt equals0I) 
  next
    assume "A = {}"
    then show "\<forall>x. \<Psi> A x = 0" unfolding \<Psi>_def by auto
  qed

lemma featfun_lm2: "(\<forall>x. \<Psi> A x = 1) = (A = UNIV)"
  proof 
    assume a0: "\<forall>x. \<Psi> A x = 1"
    then show "A = UNIV" unfolding \<Psi>_def by (smt iso_tuple_UNIV_I set_eq_iff) 
  next
    assume "A = UNIV"
    then show "\<forall>x. \<Psi> A x = 1" unfolding \<Psi>_def by auto
  qed

lemma featfun_lm3: "(\<forall>x. \<Psi> A x \<le> \<Psi> B x) = (A \<subseteq> B)" 
  sorry (* please prove it by youself *)

lemma featfun_lm4: "(\<forall>x. \<Psi> A x = \<Psi> B x) = (A = B)" 
  sorry (* please prove it by youself *)

lemma featfun_lm5: "\<Psi> (A \<inter> B) x = \<Psi> A x * \<Psi> B x" 
  sorry (* please prove it by youself *)

lemma featfun_lm6: "\<Psi> (A \<union> B) x = \<Psi> A x + \<Psi> B x - \<Psi> (A \<inter> B) x" 
  sorry (* please prove it by youself *)

lemma featfun_lm7: "\<Psi> (- A) x = 1 - \<Psi> A x" 
  sorry (* please prove it by youself *)

lemma featfun_lm8: "\<Psi> (A - B) x = \<Psi> A x - \<Psi> (A \<inter> B) x" 
  sorry (* please prove it by youself *)

lemma featfun_lm9: "\<Psi> (A - B) x = \<Psi> (A \<inter> - B) x" 
  sorry (* please prove it by youself *)

subsection {* example 7.25 *}

lemma "A \<inter> (B \<union> C) = (A \<inter> B) \<union> (A \<inter> C)"
  proof -
    {
      fix x
      have "\<Psi> (A \<inter> (B \<union> C)) x = \<Psi> A x * \<Psi> (B \<union> C) x" 
        using featfun_lm5 by force
      then have "\<Psi> (A \<inter> (B \<union> C)) x = \<Psi> A x * (\<Psi> B x + \<Psi> C x - \<Psi> (B \<inter> C) x)"
        using featfun_lm6 by force
      then have "\<Psi> (A \<inter> (B \<union> C)) x = \<Psi> A x * \<Psi> B x + \<Psi> A x * \<Psi> C x - \<Psi> A x * \<Psi> (B \<inter> C) x"
        by (simp add: featfun_lm5 featfun_lm6 inf_sup_aci(2) inf_sup_aci(3) inf_sup_distrib1)
      then have "\<Psi> (A \<inter> (B \<union> C)) x = \<Psi> (A \<inter> B) x + \<Psi> (A \<inter> C) x - \<Psi> (A \<inter> B \<inter> C) x"
        using featfun_lm5 by (metis inf_sup_aci(2)) 
      then have "\<Psi> (A \<inter> (B \<union> C)) x = \<Psi> ((A \<inter> B) \<union> (A \<inter> C)) x"
        using featfun_lm6 by (simp add: Int_Un_distrib) 
    }
    then show ?thesis using featfun_lm4 by auto
  qed

lemma "- (- A) = A"
  sorry (* please prove it by youself *)

end